#include "include.h"

using namespace std;

// globals
rgb_t vram[CFAR_IMG_WIDTH * CFAR_IMG_HEIGHT]; // hold image buffer (pixels RGBA)
unsigned int cxClient;
unsigned int cyClient;
#define WM_TICK 1

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

void set_pixel(int x, int y, rgb_t value)
{
	if (x >= CFAR_IMG_WIDTH || x < 0)
		return;
	if (y >= CFAR_IMG_HEIGHT || y < 0)
		return;

	vram[x + CFAR_IMG_WIDTH * y].r = value.r;
	vram[x + CFAR_IMG_WIDTH * y].g = value.g;
	vram[x + CFAR_IMG_WIDTH * y].b = value.b;
}

void add_pixel(int x, int y, rgb_t value)
{
	if (x >= CFAR_IMG_WIDTH || x < 0)
		return;
	if (y >= CFAR_IMG_WIDTH || y < 0)
		return;

	if (vram[x + CFAR_IMG_WIDTH * y].r == 0 &&
		vram[x + CFAR_IMG_WIDTH * y].g == 0 &&
		vram[x + CFAR_IMG_WIDTH * y].b == 0)
	{
		return;
	}

	int r = vram[x + CFAR_IMG_WIDTH * y].r;
	int g = vram[x + CFAR_IMG_WIDTH * y].g;
	int b = vram[x + CFAR_IMG_WIDTH * y].b;

	r -= value.r;
	g -= value.g;
	b -= value.b;

	if (r < 0)
		r = 0;
	if (g < 0)
		g = 0;
	if (b < 0)
		b = 0;

	vram[x + CFAR_IMG_WIDTH * y].r = r;
	vram[x + CFAR_IMG_WIDTH * y].g = g;
	vram[x + CFAR_IMG_WIDTH * y].b = b;

}

void clear_pixels()
{
	memset(vram, ~0, CFAR_IMG_WIDTH * CFAR_IMG_HEIGHT * sizeof(int));
}

uint32_t flipBytes(uint32_t n)
{
	uint32_t b0, b1, b2, b3;

	b0 = (n & 0x000000ff) << 24u;
	b1 = (n & 0x0000ff00) << 8u;
	b2 = (n & 0x00ff0000) >> 8u;
	b3 = (n & 0xff000000) >> 24u;

	return (b0 | b1 | b2 | b3);

}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	PSTR szCmdLine, int iCmdShow)
{
	static TCHAR szAppName[] = TEXT("NeuralNetwork");
	HWND         hwnd;
	MSG          msg;
	WNDCLASS     wndclass;

	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = szAppName;

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, TEXT("Program requires Windows NT!"),
			szAppName, MB_ICONERROR);
		return 0;
	}

	hwnd = CreateWindow(szAppName, TEXT("Neural Network Number Recognition"),
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, 200,
		CW_USEDEFAULT, 250,
		NULL, NULL, hInstance, NULL);

	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);

	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}

void draw_pixels(HDC hdc, HDC hdcMem, int width, int height)
{
	HBITMAP hBitmap, hOldBitmap;

	hBitmap = CreateCompatibleBitmap(hdc, width, height);
	SetBitmapBits(hBitmap, sizeof(int) * width * height, vram);
	hdcMem = CreateCompatibleDC(hdc);
	hOldBitmap = (HBITMAP)SelectObject(hdcMem, hBitmap);

	// This scaling is a little strange because Stretch maintains aspect ratios
	StretchBlt(hdc, 0, 0, cxClient, cyClient, hdcMem, 0, 0, width, height, SRCCOPY);
	SelectObject(hdcMem, hOldBitmap);
	DeleteDC(hdcMem);
	DeleteObject(hBitmap);
}



int load_random()
{
	uint8_t* train_image;
	unsigned int size;
	int selected_label = -1;

	train_image = (uint8_t*)get_file("data/train.bin", &size);

	// one byte label, 3072 byte image, repeat
	uint32_t case_count;

	case_count = size / (2 + 3072);

	int sample = rand() % case_count;

	for (int i = sample - 1; i < sample; i++)
	{
		uint8_t* label1 = train_image + 0 + i * (CFAR_IMG_WIDTH * CFAR_IMG_HEIGHT * 3 + 2);
		uint8_t* label2 = train_image + 1 + i * (CFAR_IMG_WIDTH * CFAR_IMG_HEIGHT * 3 + 2);
		uint8_t* img    = train_image + 2 + i * (CFAR_IMG_WIDTH * CFAR_IMG_HEIGHT * 3 + 2);
		selected_label = *label2;

		for (int y = 0; y < CFAR_IMG_HEIGHT; y++)
		{
			for (int x = 0; x < CFAR_IMG_WIDTH; x++)
			{
				vram[x + y * CFAR_IMG_WIDTH].r = img[x + y * CFAR_IMG_WIDTH];
				vram[x + y * CFAR_IMG_WIDTH].g = img[x + y * CFAR_IMG_WIDTH + 1024];
				vram[x + y * CFAR_IMG_WIDTH].b = img[x + y * CFAR_IMG_WIDTH + 2048];
			}
		}
	}
	
	delete[] train_image;
	return selected_label;
}



char superclass [20][80] = {
	"aquatic mammals",
	"fish",
	"flowers",
	"food containers",
	"fruit and vegetables",
	"household electrical devices",
	"household furniture",
	"insects",
	"large carnivores",
	"large man - made outdoor things",
	"large natural outdoor scenes",
	"large omnivores and herbivores",
	"medium - sized mammals",
	"non - insect invertebrates",
	"people",
	"reptiles",
	"small mammals",
	"trees",
	"vehicles 1",
	"vehicles 2"
};

char mapping[20][5][80] = {
	{"beaver", "dolphin", "otter", "seal", "whale"},
	{"aquarium fish", "flatfish", "ray", "shark", "trout"},
	{"orchids", "poppies", "roses", "sunflowers", "tulips"},
	{"bottles", "bowls", "cans", "cups", "plates"},
	{"apples", "mushrooms", "oranges", "pears", "sweet peppers"},
	{"clock", "computer keyboard", "lamp", "telephone", "television"},
	{"bed", "chair", "couch", "table", "wardrobe"},
	{"bee", "beetle", "butterfly", "caterpillar", "cockroach"},
	{"bear", "leopard", "lion", "tiger", "wolf"},
	{"bridge", "castle", "house", "road", "skyscraper"},
	{"cloud", "forest", "mountain", "plain", "sea"},
	{"camel", "cattle", "chimpanzee", "elephant", "kangaroo"},
	{"fox", "porcupine", "possum", "raccoon", "skunk"},
	{"crab", "lobster", "snail", "spider", "worm"},
	{"baby", "boy", "girl", "man", "woman"},
	{"crocodile", "dinosaur", "lizard", "snake", "turtle"},
	{"hamster", "mouse", "rabbit", "shrew", "squirrel"},
	{"maple", "oak", "palm", "pine", "willow"},
	{"bicycle", "bus", "motorcycle", "pickup truck", "train"},
	{"lawn - mower", "rocket", "streetcar", "tank", "tractor"}
};


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static RECT	rect = { 0, 0, 1, 1 };
	static HDC	hdcMem;
	PAINTSTRUCT ps;
	HDC hdc;
	static int	old_style = 0;
	static int	new_style = WS_CHILD | WS_VISIBLE;
	static int	xres, yres;
	static int best_guess[2] = { -1, -1 };
	static float best_confidence = 0.0f;
	static vector<layer_t*> layers[20]; // 20 neural networks one per superclass
	static float results[NUMBER_OF_OUTPUT_CELLS] = { 0 };
	static bool loaded = false;
	static int number = 0;



	switch (message)
	{
	case WM_TIMER:
	{
		int width = cxClient;
		int height = cyClient;

		if (!loaded)
		{
			//		printf("Training Neural Network\r\n");
			for (int i = number; i < 20; i++)
			{
				train(layers[i], i);
				number++;
				break;
			}
			//		printf("Done\r\n");

			if (number == 19)
			{
				loaded = true;
			}
		}

		InvalidateRect(hwnd, &rect, FALSE);
		return 0;
	}
	case WM_CREATE:
	{
		memset(vram, ~0, CFAR_IMG_WIDTH * CFAR_IMG_HEIGHT * sizeof(int));
		xres = CFAR_IMG_WIDTH;
		yres = CFAR_IMG_HEIGHT;

		SetWindowPos(hwnd, 0, 200, 200, 200, 250, 0);
		SetTimer(hwnd, WM_TICK, 16, NULL);
		return 0;
	}
	case WM_SIZE:
		cxClient = LOWORD(lParam);
		cyClient = HIWORD(lParam);
		rect.right = cxClient;
		rect.bottom = cyClient;
		return 0;

	case WM_KEYDOWN:
		switch (wParam)
		{
		case VK_SPACE:
			clear_pixels();
			break;
		case VK_RETURN:
		{
			best_confidence = 0.0f;

			best_guess[0] = -1;
			best_guess[1] = -1;

			for (int i = 0; i < 20; i++)
			{
				int guess = test(layers[i], vram, results, 0);

				if (results[guess] > best_confidence)
				{
					best_confidence = results[guess];
					best_guess[0] = i;
					best_guess[1] = guess;
				}
			}
			break;
		}
		case VK_TAB:
			load_random();
			break;

		}
		return 0;

	case WM_PAINT:
	{
		char guess_str[512];

		if (loaded)
		{
			if (best_guess[0] != -1)
				sprintf(guess_str, "results: %s %3.1f confidence", mapping[best_guess[0]][best_guess[1]], best_confidence);
			else
				sprintf(guess_str, "results: ? 0.0 confidence");
		}
		else
		{
			sprintf(guess_str, "Loading Training data %d of %d", number + 1, 20);
		}

		hdc = BeginPaint(hwnd, &ps);
		draw_pixels(hdc, hdcMem, xres, yres);
		DrawText(hdc, guess_str, -1, &rect, 0);
		EndPaint(hwnd, &ps);
		return 0;
	}
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}


char *get_file(char *filename, unsigned int *size)
{
	FILE	*file;
	char	*buffer;
	int	file_size, bytes_read;

	file = fopen(filename, "rb");
	if (file == NULL)
		return 0;
	fseek(file, 0, SEEK_END);
	file_size = ftell(file);
	fseek(file, 0, SEEK_SET);
	buffer = new char[file_size + 1];
	bytes_read = (int)fread(buffer, sizeof(char), file_size, file);
	if (bytes_read != file_size)
	{
		delete[] buffer;
		fclose(file);
		return 0;
	}
	fclose(file);
	buffer[file_size] = '\0';

	if (size != NULL)
	{
		*size = file_size;
	}

	return buffer;
}

uint32_t byteswap_uint32(uint32_t a)
{
	return ((((a >> 24) & 0xff) << 0) |
		(((a >> 16) & 0xff) << 8) |
		(((a >> 8) & 0xff) << 16) |
		(((a >> 0) & 0xff) << 24));
}